package rayTrace;



public class Dish extends Element {
	
	private double dishDiameter;
	private double radiusOfCurv;
	private double dishNormal[];
	private double thetaCrit;
	private double curvCentre[];
	private double reHitTolerence;
	
	public Dish(double centre[], double dishCentreNormal[], double radiusOfCurvature, double dishDiameter, int matType, double frontRIndex, double backRIndex, double reHitTolerence) {
		this.centre = centre;
		this.dishNormal = dishCentreNormal;
		this.radiusOfCurv = radiusOfCurvature;
		this.dishDiameter = dishDiameter;
		this.reHitTolerence = reHitTolerence;		
		this.matType = matType;
		this.rIndexRatio = frontRIndex / backRIndex;
		
		calcThetaCrit();
		calcCurvCentre();
	}
	
	private void calcThetaCrit(){
		thetaCrit = Math.asin( dishDiameter / (2 * radiusOfCurv ) ); //angle away from axis of dish edge
	}
	
	private void calcCurvCentre(){
		curvCentre = new double[3];		
		for(int i=0;i<3;i++)
			curvCentre[i] = centre[i] + (radiusOfCurv * dishNormal[i]); //centre of cruvature
	}
	
	public double boundingSphereRadius(){
		return Math.sqrt( Math.pow(radiusOfCurv * (1 - Math.cos(thetaCrit)),2) + dishDiameter*dishDiameter/4 );
	}
	
	public void rotate(double point[], double matrix[][]){
		for(int i=0;i<3;i++){
			centre[i] -= point[i];
			curvCentre[i] -= point[i];
		}
			
		double newCentre[] = new double[3];
		double newCurvCentre[] = new double[3];
		double newNormal[] = new double[3];
		
		for(int i=0;i<3;i++)
			for(int j=0;j<3;j++){
				newCentre[i] += matrix[i][j] * centre[j];
				newCurvCentre[i] += matrix[i][j] * curvCentre[j];
				newNormal[i] += matrix[i][j] * dishNormal[j];
			}
				
		for(int i=0;i<3;i++){
			centre[i] = point[i] + newCentre[i];
			curvCentre[i] = point[i] + newCurvCentre[i];
		}
		
		dishNormal = newNormal;
	}
	
	@Override
	public void shift(double[] dX) {
		for(int i=0;i<3;i++){
			centre[i] += dX[i];
			curvCentre[i] += dX[i];
		}
	}

	@Override
	public Intersection findIntersection(Ray ray) {
		double a=1,b=0,c=0;
		
		for(int i=0;i<3;i++){
			b += 2 * (ray.start[i] - curvCentre[i]) * ray.unit[i];
			c += ((ray.start[i] - curvCentre[i])*(ray.start[i] - curvCentre[i]));
		}
		c -= radiusOfCurv * radiusOfCurv;

		if( (b*b) < (4*a*c) ) //no roots = no intersection of sphere		
			return null;
		
		double s1,s2;
		
		//two roots give two intersection points with sphere
		s1 = ( -b - Math.sqrt(b*b - 4*a*c) ) / (2*a);
		s2 = ( -b + Math.sqrt(b*b - 4*a*c) ) / (2*a);
		
		double X1[] = new double[3];
		double X2[] = new double[3];
		double N1[] = new double[3];
		double N2[] = new double[3];
		
		for(int i=0;i<3;i++){
			X1[i] = ray.start[i] + s1 * ray.unit[i];
			X2[i] = ray.start[i] + s2 * ray.unit[i];
			N1[i] = curvCentre[i] - X1[i];
			N2[i] = curvCentre[i] - X2[i];
		}
		
		double lenN1=0,lenN2=0; 
		for(int i=0;i<3;i++){
			lenN1 += N1[i]*N1[i];
			lenN2 += N2[i]*N2[i];			
		}
		lenN1 = Math.sqrt(lenN1);
		lenN2 = Math.sqrt(lenN2);
		
		for(int i=0;i<3;i++){
			N1[i] = N1[i] / lenN1; 
			N2[i] = N2[i] / lenN2; 
		}
			
		double nnDotProd1=0, nnDotProd2=0;
		
		for(int i=0;i<3;i++){
			nnDotProd1 += dishNormal[i] * N1[i];
			nnDotProd2 += dishNormal[i] * N2[i];
		}
			
		//angle from dish normal axis 
		double theta1 = Math.acos( nnDotProd1 );
		double theta2 = Math.acos( nnDotProd2 );
	
		//which of those points are inside the dish's diameter? (& not backwards)
		boolean hit1 = (theta1 <= thetaCrit) & (s1 > reHitTolerence);
		boolean hit2 = (theta2 <= thetaCrit) & (s2 > reHitTolerence);
		
		Intersection hit = new Intersection();
	
		if( hit1){ 
			if(hit2){	//both hit, return the earliest
				if(s1 < s2){ 
					hit.pos = X1; hit.normal = N1; hit.distAlongRay = s1; hit.object = this;
				}else{
					hit.pos = X2; hit.normal = N2; hit.distAlongRay = s2; hit.object = this;
				}
				
			}else{ //only #1 hit
				hit.pos = X1; hit.normal = N1; hit.distAlongRay = s1; hit.object = this;
			}
			
		}else{
			if(hit2){ //only #2 hit
				hit.pos = X2; hit.normal = N2; hit.distAlongRay = s2; hit.object = this;
			}else{ //none hit the dish
				hit = null;
			}	
		}	 
		
		
		if(hit!=null){
			double dotProd=0;		
			for(int i=0;i<3;i++)
				dotProd += hit.normal[i] * ray.unit[i];
			
			if(dotProd < 0){ //the ray hit the front face of the dish
				hit.rIndexRatio = rIndexRatio;			
			}else{ //it hit the back face of the dish
				for(int i=0;i<3;i++) //flip the normal around, so its normal to the hitting surface
					hit.normal[i] = - hit.normal[i];
				hit.rIndexRatio = 1.0 / rIndexRatio; //and the ratio index is reversed
			}
		}
		
		return hit;
	}

	public double[] getCurvCentre() {	return curvCentre; }
	public double getDishDiameter() { return dishDiameter; }
	public double[] getDishNormal() { return dishNormal; }
	public double getRadiusOfCurv() { return radiusOfCurv; }
	public double getThetaCrit() { return thetaCrit; }


	public void setDishNormal(double[] dishNormal) {
		this.dishNormal = dishNormal;
		calcCurvCentre();
	}


	public void setRadiusOfCurv(double radiusOfCurv) {
		this.radiusOfCurv = radiusOfCurv;
		calcCurvCentre();
	}
	
	 public void setCentre(double centre[]){
		 this.centre = centre;
		 calcCurvCentre();
	 }


	

	/*
	 * //find first intersection of a ray and a dish (small part of a sphere)
		function [xPos,xNorm,dist] = dishIntersect(ray, dish)
			thetaCrit = asin( dish.diameter / (2 * dish.radius ) ); //angle away from axis of dish edge
		
			curvCentre = dish.centre + (dish.radius .* dish.normal); //centre of cruvature
		
			a = 1; 
			b = 2 * sum( (ray.start - curvCentre) .* ray.unit );
			c = sum( (ray.start - curvCentre)^2 ) - dish.radius^2;
		
			if (b^2) < (4*a*c) //no roots = no intersection of sphere		
				xNorm = [0,0,0]; xPos=[0,0,0]; dist=0;
				return;
			end;
			
			//two roots give two intersection points with sphere
			s1 = ( -b - sqrt(b^2 - 4*a*c) ) / (2*a);
			s2 = ( -b + sqrt(b^2 - 4*a*c) ) / (2*a);
			X1 = ray.start + s1 * ray.unit;
			X2 = ray.start + s2 * ray.unit;
		
			//& the normals to the surface at those points
			N1 = curvCentre - X1; N1 = N1 ./ sqrt(sum(N1^2));
			N2 = curvCentre - X2; N2 = N2 ./ sqrt(sum(N2^2));
		
			//angle from dish normal axis 
			theta1 = acos( sum( dish.normal .* N1 ));
			theta2 = acos( sum( dish.normal .* N2 ));
		
			//which of those points are inside the dish's diameter? (& not backwards)
			hit1 = (theta1 <= thetaCrit) & (s1 > 0);
			hit2 = (theta2 <= thetaCrit) & (s2 > 0);
		
			if hit1 
				if hit2	//both hit, return the earliest
					if s1 < s2 
						xPos = X1; xNorm = N1; dist=s1;
					else
						xPos = X2; xNorm = N2; dist=s2;
					end
				else //only #1 hit
					xPos = X1; xNorm = N1; dist=s1;
				end
			else
				if hit2 //only #2 hit
					xPos = X2; xNorm = N2; dist=s2;
				else //none hit the dish
					xNorm = [0,0,0]; xPos=[0,0,0]; dist=0;
				end	
			end;
		endfunction 
	 */

}
